import sys
if sys.version_info.major == 3:
    from configparser import ConfigParser
else:
    input = raw_input
    from ConfigParser import ConfigParser
from datetime import datetime
import logging
import os
from os import path
import re
import shutil
import subprocess

from splunk.clilib.bundle_paths import make_splunkhome_path


FAILURE_FLAG = False
FAILURE_MESSAGES = []
CURRENT_TIME = datetime.today().strftime('%d%m%y-%H%M%S')

COLLECTION_CONF_PATH = make_splunkhome_path([
    'etc', 'apps', 'Splunk_TA_vmware_inframon', 'local', 'inframon_ta_vmware_collection.conf'])

TEMPLATE_CONF_PATH = make_splunkhome_path([
    'etc', 'apps', 'Splunk_TA_vmware_inframon', 'local', 'inframon_ta_vmware_template.conf'])

INPUTS_CONF_PATH = make_splunkhome_path([
    'etc', 'apps', 'Splunk_TA_vmware_inframon', 'local', 'inputs.conf'])

COLLECTION_CONF_BACKUP_PATH = make_splunkhome_path([
    'etc', 'apps', 'Splunk_TA_vmware_inframon', 'local', 'inframon_ta_vmware_collection.conf_' + CURRENT_TIME + '.bak'])

TEMPLATE_CONF_BACKUP_PATH = make_splunkhome_path([
    'etc', 'apps', 'Splunk_TA_vmware_inframon', 'local', 'inframon_ta_vmware_template.conf_' + CURRENT_TIME + '.bak'])

RENAME_DICT = {
    'managed_host_whitelist': 'managed_host_includelist',
    'managed_host_blacklist': 'managed_host_excludelist',
    'vm_instance_whitelist': 'vm_instance_allowlist',
    'vm_instance_blacklist': 'vm_instance_denylist',
    'host_instance_whitelist': 'host_instance_allowlist',
    'host_instance_blacklist': 'host_instance_denylist',
    'cluster_instance_whitelist': 'cluster_instance_allowlist',
    'cluster_instance_blacklist': 'cluster_instance_denylist',
    'perf_entity_blacklist': 'perf_entity_denylist',
}


def setup_logger(logger_name):
    """
    Creates and returns the logger object with the given name
    :logger_name: Name of the logger
    :return: logger object
    """
    logging.basicConfig(filename=make_splunkhome_path(['var', 'log', 'splunk', 'upgrade_metric_TA_from_410_to_420.log']),
                        format='%(asctime)s %(levelname)s %(lineno)d - %(message)s',
                        filemode='a')
    logger = logging.getLogger(logger_name)
    logger.setLevel(logging.INFO)
    return logger


logger = setup_logger('upgrade_metric_TA')


def add_failure(message=''):
    """
    Updates the global Failure Flag and add the global Failure Message list
    :message: The mesage to add in the failure messages
    :return: nothing
    """
    global FAILURE_FLAG
    FAILURE_FLAG = True
    if message:
        FAILURE_MESSAGES.append(message)


def log_and_add_failure(error_message, exception):
    """
    Logs the error in the log file and add the failure message to global Failure message list
    :error_message: Message to add in the global failure message list
    :exception: Exception object
    """
    logger.error("{} Error: {}".format(error_message, exception))
    add_failure(error_message)


def directory_exists(*directories):
    """
    Checks the existence of the package in etc/apps
    :param directories: dynamic argument of sequential names of directories
    :return: Boolean based on existance of package
    """
    DIR_PATH = ['etc', 'apps'] + [directory for directory in directories]
    return path.isdir(make_splunkhome_path(DIR_PATH))


def get_user_permission_to_move_ahead(message):
    """
    Called when the script is not able to determine the current Splunk status.
    :message: The message to show to user before prompting for the choice
    :return: True if the user opts to move ahead, else False.
    """
    try:
        print(message)
        print("Enter\n1) To proceed ahead\nAny other key to exit:")

        user_choice = input()
        user_choice = int(user_choice)

        if user_choice == 1:
            logger.info("User selected to proceed ahead.")
            return True
        else:
            print(
                "Terminating the script run, please run the upgrade script after stopping the Splunk.")
            logger.info(
                "User terminated the script.")
            sys.exit(0)
    except ValueError as e:
        error_message = "Invalid value entered. Terminating the script run, please run the upgrade script after stopping the Splunk."
        print(error_message)
        logger.error("{} Error: {}".format(error_message, e))
        sys.exit(0)
    except Exception as e:
        error_message = "Error occured while getting the user input. Terminating the upgrade script run."
        print(error_message)
        logger.error("{} Error: {}".format(error_message, e))
        sys.exit(0)


def start_splunk():
    """
    Starts the Splunk
    :return: Boolean variable denoting the operation status
    """
    try:
        output_message, output_error = run_splunk_command('start')
        logger.info(
            "Output message for Start command: {}".format(output_message))
        logger.info("Output error for Start command: {}".format(output_error))

        if((not output_error) and (b"The Splunk web interface is at" in output_message)):
            logger.info("Splunk is successfully started.")
            print("Splunk is successfully started.")
            return
    except Exception as e:
        logger.error(
            "Error occured while starting the Splunk. Error: {}".format(e))

    print("Error occured while starting the Splunk. Please start the Splunk manually.")


def stop_splunk():
    """
    Stops the Splunk
    :return: Boolean variable denoting the operation status
    """
    try:
        output_message, output_error = run_splunk_command('stop')
        logger.info(
            "Output message for Stop command: {}".format(output_message))
        logger.info("Output error for Stop command: {}".format(output_error))

        if((not output_error) and (b"Done." in output_message) and (not is_splunk_running())):
            logger.info("Splunk stopped successfully.")
            print("Splunk stopped successfully.")
            return True
    except Exception as e:
        logger.error(
            "Error occured while stopping the Splunk. Error: {}".format(e))

    error_message = "Error occured while stopping the Splunk, please make sure Splunk is stopped before upgrading."
    return get_user_permission_to_move_ahead(error_message)


def is_splunk_running():
    """
    Checks the Splunk status
    :return: Boolean representing the Splunkd status by running the `splunk status` command by spawning a subprocess.
    """
    try:
        output_message, output_error = run_splunk_command('status')
        logger.info(
            "Output message for status command: {}".format(output_message))
        logger.info("Output error for status command: {}".format(output_error))

        if output_error:
            error_message = "Unable to determine status of Splunk. Please make sure Splunk is stopped before upgrading."
            logger.error("{} Error: {}".format(output_message, output_error))
            # If user agrees to move ahead, then that means that Splunk is stopped, returning False
            return not get_user_permission_to_move_ahead(error_message)
        else:
            if b"splunkd is running" in output_message:
                logger.info("Splunk service is running.")
                return True
            else:
                logger.info("Splunk service is not running.")
                return False
    except Exception as e:
        error_message = "Error occured while determining status of Splunk. Please make sure Splunk is stopped before proceeding ahead."
        logger.error("{} Error: {}".format(error_message, e))

        return not get_user_permission_to_move_ahead(error_message)


def run_splunk_command(command):
    """
    Runs the specified Splunk command from command-line.
    :command: Name of the Splunk command that needs to be run
    :return: error and the message returned by communicate method of subprocess.Popen
    """
    splunk_cmd = make_splunkhome_path(['bin', 'splunk'])

    output = subprocess.Popen([splunk_cmd, command],
                              stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
    output_message, output_error = output.communicate()
    return output_message, output_error


def start_splunk_wizard():
    """
    Starts the splunk service after obtaining user's choice
    :return: Boolean indicating the status of the Splunk
    """
    print("\nEnter 1 to start the Splunk or enter any other key to skip starting the Splunk:")
    try:
        user_choice = input()
        user_choice = int(user_choice)

        if user_choice == 1:
            print("Starting the Splunk on this instance...")
            logger.info("Starting the Splunk on this instance...")
            start_splunk()
        else:
            raise ValueError()
    except ValueError:
        print("Existing the upgrade script, please start the Splunk manually.")
        logger.error(
            "Existing the upgrade script, please start the Splunk manually.")
    except Exception as e:
        print(
            "Error occured while getting the user input, please start the Splunk manually.")
        logger.error(
            "Error occured while getting the user input, please start the Splunk manually. Error: {}".format(e))


def stop_splunk_wizard():
    """
    Stops the splunk service after obtaining user's choice
    :return: Boolean indicating whether to continue operation or abort it.
    """
    print("\nSplunk needs to be stopped before running the upgrade script.\nEnter 1 to stop the Splunk or Enter any other key to terminate upgrade script:")
    try:
        user_choice = input()
        user_choice = int(user_choice)

        if user_choice == 1:
            print("Stopping Splunk on this instance...")
            logger.info("Stopping Splunk on this instance...")
            return stop_splunk()
        else:
            raise ValueError()
    except ValueError as e:
        print(
            "Exiting the upgrade script, please stop the Splunk and run the upgrade script again.")
        logger.error(
            "Exiting the upgrade script, please stop the Splunk and run the upgrade script again.")
        return False
    except Exception as e:
        print("Error occured while getting input, please stop the Splunk and run the upgrade script again.")
        logger.error(
            "Error occured while getting input, please stop the Splunk and run the upgrade script again. Error: {}".format(e))
        return False


def run_prechecks():
    """
    Run the prechecks for the upgrade script
    :return: boolean denoting whether to do upgrade or not.
    """
    if is_splunk_running():
        if not stop_splunk_wizard():
            return False

    return True


def write_conf_file(conf_path, config):
    """
    Dumps content of config in configuration file present in conf_path
    :conf_path: absolute path of the configuration file
    :config: the object of ConfigParser class having content to write
    """
    try:
        # Writing the configuration file
        with open(conf_path, 'w') as file_object:
            config.write(file_object)
            logger.info(
                "Successfully updated configuration file present at: {}".format(conf_path))
    except IOError as e:
        error_message = "Error occured while writing configuration file present at {}".format(
            conf_path)
        log_and_add_failure(error_message, e)
    except Exception as e:
        error_message = "Error occured while writing configuration file present at {}".format(
            conf_path)
        log_and_add_failure(error_message, e)


def data_collection_selection_wizard():
    """
    Presents the selection menu to user for whether to start data collection or not
    :return: Boolean indicating whether to show message to user to enable the data collection from UI
    """
    print("\nEnter 1 to enable the Data Collection inputs or enter any other key to skip enabling:")
    try:
        user_choice = input()
        user_choice = int(user_choice)

        if user_choice == 1:
            print("Enabling the data collection inputs...")
            logger.info("Adding and enabling the data collection inputs...")
            return create_and_update_inputs()
        else:
            raise ValueError()
    except ValueError:
        print("Skipping enabling the data collection...")
        logger.error(
            "Skipping enabling the data collection...")
        return False
    except Exception as e:
        print("Error occured while getting user input.")
        logger.error(
            "Error occured while getting user input in the data collection wizard, Error: {}".format(e))
        return False


def create_and_update_inputs():
    """
    Responsible for enabling the data collection inputs in the TA after successful upgrade
    """
    if directory_exists('Splunk_TA_vmware_inframon', 'local'):
        try:

            inputs_config = ConfigParser()
            if path.exists(INPUTS_CONF_PATH):
                inputs_config.read(INPUTS_CONF_PATH)

            input_names = [
                'script://$SPLUNK_HOME/etc/apps/Splunk_TA_vmware_inframon/bin/ta_vmware_hierarchy_agent.py',
                'ta_vmware_collection_scheduler_inframon://Global pool'
            ]

            logger.info("Enabling Scheduler and hierarchy agent inputs...")

            for input_name in input_names:
                if not inputs_config.has_section(input_name):
                    inputs_config.add_section(input_name)

                inputs_config.set(input_name, 'disabled', '0')

            with open(INPUTS_CONF_PATH, 'w') as file_object:
                inputs_config.write(file_object)
                logger.info(
                    "Successfully updated configuration file present at: {}".format(INPUTS_CONF_PATH))

            logger.info(
                "Successfully created inputs.conf file and enabled inputs.")
            print("Successfully enabled data collection inputs.")

            try:
                # Updating the permissions of the file
                logger.info(
                    "Setting appropriate permissions on: {}".format(INPUTS_CONF_PATH))
                os.chmod(INPUTS_CONF_PATH, 0o600)
                logger.info(
                    "Successfully set permissions on: {}".format(INPUTS_CONF_PATH))

                return True
            except Exception as e:
                error_message = "Unable to set the permissions on file present at {}".format(
                    INPUTS_CONF_PATH)
                logger.error("{}. Error: {}".format(error_message, e))
                return False

        except IOError as e:
            error_message = "Error occured while creating inputs.conf file."
            logger.error("{}. Error: {}".format(error_message, e))
            print(error_message)
            return False
        except Exception as e:
            error_message = "Error occured while creating inputs.conf file."
            logger.error("{}. Error: {}".format(error_message, e))
            print(error_message)
            return False
    else:
        error_message = "local directory is not present in Splunk_TA_vmware_inframon directory."
        logger.error(error_message)
        print(error_message)
        return False


def update_collection_stanza(stanza, config, renamed_parameters):
    """
    Updates the stanza by renaming the biased stanza params with the non-biased stanza params
    """

    logger.info(
        "Looking for the biased parameters under the {} stanza of the inframon_ta_vmware_collection.conf file.".format(stanza))
    for key, val in config.items(stanza):
        if re.search(renamed_parameters, key):
            renamed_key = RENAME_DICT.get(key, None)
            if renamed_key:
                logger.info("biased parameter {} has been found. Replacing it with:{}...".format(
                key, renamed_key))
                config.set(stanza, renamed_key, val)
                config.remove_option(stanza, key)


def update_conf_files():
    """
    Updates the inframon_ta_vmware_collection.conf file by replacing the biased param names with its respective non-biased param names.
    Removes the deprecated parameters from inframon_ta_vmware_template.conf file
    """
    global FAILURE_FLAG
    if path.exists(COLLECTION_CONF_PATH):
        try:
            # Reading the inframon_ta_vmware_collection.conf file
            config = ConfigParser()
            config.read(COLLECTION_CONF_PATH)
            renamed_parameters = re.compile(r".*_whitelist|.*_blacklist")

            # Searching the moved parameters in the 'default' stanza

            for section in config.sections():
                update_collection_stanza(section, config, renamed_parameters)

            write_conf_file(COLLECTION_CONF_PATH, config)

            print("Successfully updated inframon_ta_vmware_collection.conf file.")

        except Exception as e:
            error_message = "Error occured while updating inframon_ta_vmware_collection.conf file."
            log_and_add_failure(error_message, e)
    else:
        print('inframon_ta_vmware_collection.conf does not exist in the local directory. No need to perform any steps related to it.')
        logger.info("inframon_ta_vmware_collection.conf does not exist in the local directory. No need to perform any steps related to it.")

    if path.exists(TEMPLATE_CONF_PATH):
        try:
            # Reading inframon_ta_vmware_template.conf file and removing the deprecated parameters
            config = ConfigParser()
            config.read(TEMPLATE_CONF_PATH)
            deprecated_parameters = re.compile(r".*_perf_fields")

            # Searching for the deprecated parameters in each stanza and removingn them as they are not useful
            logger.info("Checking for deprecated parameters in inframon_ta_vmware_template.conf file...")
            for section in config.sections():
                for key, _ in config.items(section):
                    if re.search(deprecated_parameters, key):
                        logger.info("Deprecated parameter {} has been found. Removing it...".format(key))
                        config.remove_option(section, key)
                        logger.info("Depreacted parameter {} has been removed.".format(key))

            write_conf_file(TEMPLATE_CONF_PATH, config)

            print("Successfully removed deprecated parameters from inframon_ta_vmware_template.conf file.")
        except Exception as e:
            error_message = "Error occured while removing deprecated parameters inframon_ta_vmware_template.conf file."
            log_and_add_failure(error_message, e)
    else:
        print('inframon_ta_vmware_template.conf does not exist in the local directory. No need to perform any steps related to it.')
        logger.info("inframon_ta_vmware_template.conf does not exist in the local directory. No need to perform any steps related to it.")


def create_backup_file(file_path, backup_file_path):
    """
    Creates backup file of the given file in the same directory
    :file_path: Path of the file which need to be backed up
    :backup_file_path: String representing path of backup file
    :return: Boolean indicating whether operation was successful or not
    """
    global FAILURE_FLAG
    print("Taking the backup of file {} before renaming the params...".format(file_path))
    logger.info("Taking the backup of file {}".format(file_path))
    try:
        shutil.copy(file_path, backup_file_path)
        logger.info(
            "Successfully backed up the {} to {}".format(file_path, backup_file_path))
        return True
    except Exception as e:
        error_message = "Error while taking the backup of file {}".format(file_path)
        print("Error while taking the backup of {}, Aborting the upgrade. Please try running the script again.".format(file_name))
        log_and_add_failure(error_message, e)
        return False


def upgrade_scheduler():
    """
    Upgrades the instance by following upgrade steps of Scheduler
    """

    global FAILURE_FLAG

    if path.exists(COLLECTION_CONF_PATH) and not path.exists(COLLECTION_CONF_BACKUP_PATH):
        if not create_backup_file(COLLECTION_CONF_PATH, COLLECTION_CONF_BACKUP_PATH):
            sys.exit()

    if path.exists(TEMPLATE_CONF_PATH) and not path.exists(TEMPLATE_CONF_BACKUP_PATH):
        if not create_backup_file(TEMPLATE_CONF_PATH, TEMPLATE_CONF_BACKUP_PATH):
            sys.exit()

    update_conf_files()

    if FAILURE_FLAG:
        print("Errors occurred while taking the backup of the Splunk_TA_vmware_inframon and SA-Hydra-inframon packages. Please run the script again. Errors: {}".format(
            FAILURE_MESSAGES))
        sys.exit()

    print("Add-on upgrade completed successfully.")
    logger.info("Add-on upgrade completed successfully.")

    if not data_collection_selection_wizard():
        print("If you would like to start the data collection, then please go to the Collection Configuration page of Splunk Add-on for VMware Metrics and enable the data collection.")

    start_splunk_wizard()


def main():

    logger.info("Initiating the upgrade script for the Splunk Add-on for VMware Metrics...")

    if run_prechecks():
        upgrade_scheduler()


if __name__ == "__main__":
    main()
